import snowflake.connector
import pandas as pd
import os
from typing import Dict, List, Optional, Union
from shipyard_templates import Database, ExitCodeException, ShipyardLogger
from snowflake.connector import pandas_tools as pt
from shipyard_snowflake.utils import utils
from shipyard_snowflake.utils.exceptions import (
    SnowflakeToPandasError,
    PandasToSnowflakeError,
    SchemaInferenceError,
    PutError,
    CopyIntoError,
    QueryExecutionError,
    DownloadError,
    CreateTableError,
)

logger = ShipyardLogger.get_logger()


# TODO: Refactor the exit codes and exceptions to confrom to the new standard by the other databases
class SnowflakeClient(Database):
    EXIT_CODE_PUT_ERROR = 101
    EXIT_CODE_COPY_INTO_ERROR = 102
    EXIT_CODE_SCHEMA_INFERENCE_ERROR = 103
    EXIT_CODE_PANDAS_TO_SNOWFLAKE_CONVERSION_ERROR = 104
    EXIT_CODE_SNOWFLAKE_TO_PANDAS_CONVERSION_ERROR = 105
    EXIT_CODE_CREATE_TABLE_ERROR = 106
    EXIT_CODE_DOWNLOAD_ERROR = 107
    EXIT_CODE_NON_EMPTY_TABLE = 108

    def __init__(
        self,
        username,
        password,
        database=None,
        account=None,
        warehouse=None,
        schema=None,
        rsa_key=None,
        role=None,
    ) -> None:
        self.username = username
        self.password = password
        self.account = account
        self.warehouse = warehouse
        self.schema = schema
        self.database = database
        self.rsa_key = rsa_key
        self.role = role
        self.application = os.environ.get("SHIPYARD_USER_AGENT", None)
        super().__init__(
            username,
            password,
            account=account,
            warehouse=warehouse,
            schema=schema,
            database=database,
            rsa_key=rsa_key,
            application=self.application,
        )

    # TODO: "Refactor this function to be more concise"
    def connect(self):
        """Helper function for authentication tests to see if provided credentials are valid"""
        if self.rsa_key:
            if self.password:
                logger.warning(
                    "Private Key was provided in addition to Password. Using the Private Key to login"
                )
            private_key = utils._decode_rsa(self.rsa_key)
            try:
                if not self.application:
                    con = snowflake.connector.connect(
                        user=self.username,
                        account=self.account,
                        private_key=private_key,
                        warehouse=self.warehouse,
                        database=self.database,
                        schema=self.schema,
                        role=self.role,
                    )
                else:
                    con = snowflake.connector.connect(
                        user=self.username,
                        account=self.account,
                        private_key=private_key,
                        warehouse=self.warehouse,
                        database=self.database,
                        schema=self.schema,
                        role=self.role,
                        application=self.application,
                    )
                logger.info("Successfully connected to Snowflake")
                self.conn = con
                return con
            except Exception as e:
                raise ExitCodeException(
                    message=f"Could not authenticate to Snowflake for user {self.username} with credentials in {self.rsa_key}",
                    exit_code=self.EXIT_CODE_INVALID_CREDENTIALS,
                )
        else:
            try:
                if not self.application:
                    con = snowflake.connector.connect(
                        user=self.username,
                        password=self.password,
                        account=self.account,
                        warehouse=self.warehouse,
                        database=self.database,
                        schema=self.schema,
                        role=self.role,
                    )
                else:
                    con = snowflake.connector.connect(
                        user=self.username,
                        password=self.password,
                        account=self.account,
                        warehouse=self.warehouse,
                        database=self.database,
                        schema=self.schema,
                        role=self.role,
                        application=self.application,
                    )
                logger.info("Successfully connected to snowflake")
                # TODO: Refactor this to be a property like the other databases
                self.conn = con
                return con
            except Exception as e:
                raise ExitCodeException(
                    f"Could not authenticate to Snowflake due to {e}",
                    self.EXIT_CODE_INVALID_CREDENTIALS,
                )

    def upload(
        self,
        file_path: str,
        table_name: str,
        insert_method: str = "replace",
    ):
        """Uploads a pandas dataframe to a snowflake table

        Args:
            df (pd.DataFrame): The dataframe to be uploaded
            table_name (str): The name of the Snowflake Table to, if it doesn't exist, it will be created
            insert_method (str): The method to use when inserting the data into the table. Options are replace or append Defaults to 'replace'
        """
        if insert_method not in ["replace", "append", "add"]:
            raise ExitCodeException(
                f"Invalid insert method: {insert_method} is not a valid insert method. Choose between 'replace' or 'append'",
                self.EXIT_CODE_INVALID_ARGUMENTS,
            )

        try:
            if insert_method == "add" and not self._is_empty(table_name):
                raise ExitCodeException(
                    f"Error: The table {table_name} already has data in it. Instead of selecting the `Add Data Only if Table is Empty` option, select either `Replace` or `Append`, or delete all the rows in the target table",
                    exit_code=self.EXIT_CODE_NON_EMPTY_TABLE,
                )
            self.put(file_path=file_path, table_name=table_name)
            self.copy_into(table_name=table_name, insert_method=insert_method)
        except PutError as ec:
            logger.error(ec.message)
            raise ExitCodeException(ec.message, ec.exit_code)
        except CopyIntoError as ec:
            raise ExitCodeException(ec.message, ec.exit_code)

        except ExitCodeException as ec:
            logger.error("Error in uploading file")
            raise ExitCodeException(ec.message, ec.exit_code)
        except Exception as e:
            logger.error(f"Unknown error in uploading file: {str(e)}")
            raise ExitCodeException(str(e), self.EXIT_CODE_INVALID_UPLOAD_VALUE)

    def execute_query(self, query: str) -> snowflake.connector.cursor.SnowflakeCursor:
        """Executes a query in Snowflake

        Args:
            query: The query to execute

        Returns: The cursor object

        """
        try:
            cursor = self.conn.cursor()
            res = cursor.execute(query)
            logger.debug(f"Successfully executed query: `{query}` in Snowflake")
            return res
        except Exception as e:
            raise ExitCodeException(str(e), self.EXIT_CODE_INVALID_QUERY)

    def fetch(self, query: str) -> pd.DataFrame:
        """Fetches the results of a Snowflake SQL query as a pandas dataframe

        Args:
            query: The query to send to Snowflake


        Returns: The query results as a dataframe

        """
        try:
            cursor = self.execute_query(query)
            results = cursor.fetch_pandas_all()
        except ExitCodeException as ec:
            raise DownloadError(
                f"Error in fetching query results. Message from snowflake includes: {ec.message}",
                exit_code=self.EXIT_CODE_DOWNLOAD_ERROR,
            )
        except Exception as e:
            raise DownloadError(
                f"Error in fetching query results. Message from snowflake includes: {str(e)}",
                exit_code=self.EXIT_CODE_DOWNLOAD_ERROR,
            )
        else:
            return results

    def put(
        self,
        file_path: str,
        table_name: str,
    ):
        """Executes a PUT command to load a file to internal staging. This is the fastest way to load a large file and should be followed by a copy into command

        Args:
            file_path: The file to load
            table_name: The table in Snowflake to write to
        """
        put_statement = f"""PUT file://{file_path} '@%{table_name}' OVERWRITE=TRUE """
        try:
            self.execute_query(put_statement)
        except ExitCodeException as ec:
            raise PutError(
                message=f"Error in executing PUT query. Message from snowflake includes: {ec.message}",
                exit_code=self.EXIT_CODE_PUT_ERROR,
            )

    def copy_into(self, table_name: str, insert_method: str):
        """
        Executes a COPY INTO command to load a file from internal staging to a table

        Args:
            table_name: The name of the destination table to copy into
            insert_method: Whether th replace or append to the table
        """
        copy_statement = f"""COPY INTO {table_name} FROM '@%{table_name}' PURGE=TRUE FILE_FORMAT=(TYPE=CSV FIELD_DELIMITER=',' COMPRESSION=GZIP, PARSE_HEADER=TRUE) MATCH_BY_COLUMN_NAME=CASE_INSENSITIVE"""
        if insert_method == "append":
            copy_statement += f" ON_ERROR = CONTINUE"
        try:
            self.execute_query(copy_statement)
        except ExitCodeException as ec:
            raise CopyIntoError(
                message=f"Could not execute COPY INTO statement to target table. Message from Snowflake includes: {str(ec.message)}",
                exit_code=self.EXIT_CODE_COPY_INTO_ERROR,
            )

    def _create_table_sql(
        self,
        table_name: str,
        columns: Union[List[List[str]], Dict[str, str]],
    ) -> str:
        """Returns the SQL for to create or replace a table in Snowflake
        Args:
            table_name (str): The name of the table to create
            columns (List[List[str,str]]): A list of lists of the column name and the data type. Example: [("column1","varchar(100)"),("column2","varchar(100)")]. Defaults to None
        """
        # this is for backwards compatibility
        if isinstance(columns, list):
            column_string = ",".join([f'"{col[0]}" {col[1]}' for col in columns])
        else:
            column_string = ",".join([f'"{k}" {v}' for k, v in columns.items()])

        create_statement = f"""CREATE OR REPLACE TABLE {table_name} ({column_string})"""
        return create_statement

    def create_table(self, sql: str):
        """Helper function to create a table in Snowflake

        Args:
            sql: The create statement generated by _create_table_sql

        """
        try:
            self.execute_query(query=sql)
        except ExitCodeException as ec:
            raise CreateTableError(
                f"Error in creating table. Message from Snowflake includes: {ec.message}",
                exit_code=self.EXIT_CODE_CREATE_TABLE_ERROR,
            )

    def _exists(self, table_name: str) -> bool:
        """Helper function to check if a given table exists

        Args:
            table_name (): The name of the table

        Returns: True if exists, False if not
        """
        query = f"""SELECT * FROM INFORMATION_SCHEMA.TABLES 
        WHERE TABLE_NAME = '{str(table_name).upper()}'
        """
        if self.database:
            query += f" AND TABLE_CATALOG = '{str(self.database).upper()}'"
        if self.schema:
            query += f" AND TABLE_SCHEMA = '{str(self.schema).upper()}'"

        res = self.fetch(query)

        if res.shape[0] == 1:
            return True

        return False

    def _is_empty(self, table_name: str):
        sql = f"SELECT COUNT(*) FROM {table_name}"
        cur = self.execute_query(sql)
        result = cur.fetchone()[0]
        return result == 0

    def close(self):
        """
        Closes the database connection
        """
        self.conn.close()



import os
import pandas as pd
import pytest
from shipyard_snowflake import SnowflakeClient
from dotenv import load_dotenv, find_dotenv
from shipyard_snowflake.utils.utils import (
    map_pandas_to_snowflake,
    infer_schema,
    reservoir_sample,
    read_file,
)

if env_exists := os.path.exists(".env"):
    load_dotenv()

    df_path = "shipyard_snowflake/test/simple.csv"

    user = os.getenv("SNOWFLAKE_USER")
    pwd = os.getenv("SNOWFLAKE_PWD")
    account = os.getenv("SNOWFLAKE_ACCOUNT")
    schema = os.getenv("SNOWFLAKE_SCHEMA")
    database = os.getenv("SNOWFLAKE_DATABASE")
    warehouse = os.getenv("SNOWFLAKE_WAREHOUSE")
    # rsa_key = "/Users/wespoulsen/.ssh/snowflake_key.p8"
    role = os.getenv("SNOWFLAKE_ROLE")

    client = SnowflakeClient(
        username=user,
        password=pwd,
        database=database,
        account=account,
        warehouse=warehouse,
        schema=schema,
        role=role,
    )
    snowflake_dtypes = [
        ["string_col", "VARCHAR"],
        ["char_col", "CHAR"],
        ["int_col", "INT"],
        ["float_col", "FLOAT"],
        ["date_col", "DATE"],
        ["datetime_col", "TIMESTAMP"],
        ["bool_col", "BOOLEAN"],
    ]


@pytest.mark.skipif(not env_exists, reason="No .env file found")
def test_put():
    conn = client.connect()
    create_sql = client._create_table_sql(
        table_name="PUT_TEST", columns=snowflake_dtypes
    )
    client.execute_query(conn=conn, query=create_sql)
    client.put(conn, file_path=df_path, table_name="PUT_TEST")
    client.copy_into(conn, table_name="PUT_TEST")


@pytest.mark.skipif(not env_exists, reason="No .env file found")
def test_put_no_datatypes():
    conn = client.connect()
    sample = infer_schema(df_path)
    snowflake_dtypes = map_pandas_to_snowflake(sample)
    create_sql = client._create_table_sql(
        table_name="LARGER_TEST", columns=snowflake_dtypes
    )
    client.execute_query(conn, create_sql)
    client.put(
        conn, file_path="shipyard_snowflake/test/larger.csv", table_name="LARGER_TEST"
    )
    client.copy_into(conn, table_name="LARGER_TEST")



from shipyard_templates import ExitCodeException


# These are custom exceptions specific to shipyard-snowflake
class SnowflakeToPandasError(ExitCodeException):
    def __init__(self, message: str, exit_code: int):
        super().__init__(message, exit_code)
        self.message = message
        self.exit_code = exit_code


class PandasToSnowflakeError(ExitCodeException):
    def __init__(self, message: str, exit_code: int):
        super().__init__(message, exit_code)
        self.message = message
        self.exit_code = exit_code


class SchemaInferenceError(ExitCodeException):
    def __init__(self, message: str, exit_code: int):
        super().__init__(message, exit_code)
        self.message = message
        self.exit_code = exit_code


class PutError(ExitCodeException):
    def __init__(self, message: str, exit_code: int):
        super().__init__(message, exit_code)
        self.message = message
        self.exit_code = exit_code


class CopyIntoError(ExitCodeException):
    def __init__(self, message: str, exit_code: int):
        super().__init__(message, exit_code)
        self.message = message
        self.exit_code = exit_code


class QueryExecutionError(ExitCodeException):
    def __init__(self, message: str, exit_code: int):
        super().__init__(message, exit_code)
        self.message = message
        self.exit_code = exit_code


class DownloadError(ExitCodeException):
    def __init__(self, message: str, exit_code: int):
        super().__init__(message, exit_code)
        self.message = message
        self.exit_code = exit_code


class CreateTableError(ExitCodeException):
    def __init__(self, message: str, exit_code: int):
        super().__init__(message, exit_code)
        self.message = message
        self.exit_code = exit_code



import os
import pandas as pd
import pytest
from shipyard_snowflake import SnowparkClient
from dotenv import load_dotenv, find_dotenv

if env_exists := os.path.exists(".env"):
    load_dotenv()

    df_path = "shipyard_snowflake/test/simple.csv"

    user = os.getenv("SNOWFLAKE_USER")
    pwd = os.getenv("SNOWFLAKE_PWD")
    account = os.getenv("SNOWFLAKE_ACCOUNT")
    schema = os.getenv("SNOWFLAKE_SCHEMA")
    database = os.getenv("SNOWFLAKE_DATABASE")
    warehouse = os.getenv("SNOWFLAKE_WAREHOUSE")
    # rsa_key = "/Users/wespoulsen/.ssh/snowflake_key.p8"
    role = os.getenv("SNOWFLAKE_ROLE")

    client = SnowparkClient(
        username=user,
        pwd=pwd,
        database=database,
        account=account,
        warehouse=warehouse,
        schema=schema,
    )


@pytest.mark.skipif(not env_exists, reason="No .env file found")
def test_upload():
    session = client.connect()
    df = pd.read_csv(df_path)
    client.upload(session=session, df=df, table_name="SNOWPARK_TEST", overwrite=True)
    print("Done")


@pytest.mark.skipif(not env_exists, reason="No .env file found")
def test_larger_upload():
    session = client.connect()
    larger = "shipyard_snowflake/test/larger.csv"
    df = pd.read_csv(larger)
    client.upload(session=session, df=df, table_name="LARGER_TEST", overwrite=True)
    print("Done")


@pytest.mark.skipif(not env_exists, reason="No .env file found")
def test_larger_upload_append():
    session = client.connect()
    larger = "shipyard_snowflake/test/larger.csv"
    df = pd.read_csv(larger)
    client.upload(session=session, df=df, table_name="LARGER_TEST", overwrite=False)
    print("Done")


@pytest.mark.skipif(not env_exists, reason="No .env file found")
def test_put():
    session = client.connect()
    larger = "shipyard_snowflake/test/larger.csv"
    client.put(
        session=session, file_path=larger, table_name="LARGER_TEST", overwrite=True
    )
    print("Done")
    # client.copy_into(session=session, table_name="LARGER_TEST", overwrite=True)



import argparse
import sys
from shipyard_snowflake import SnowflakeClient
from shipyard_templates import ExitCodeException, ShipyardLogger

logger = ShipyardLogger.get_logger()


def get_args():
    parser = argparse.ArgumentParser()
    parser.add_argument("--username", dest="username", required=True)
    parser.add_argument("--password", dest="password", required=False, default="")
    parser.add_argument(
        "--private-key-path", dest="private_key_path", required=False, default=""
    )
    parser.add_argument(
        "--private-key-passphrase",
        dest="private_key_passphrase",
        required=False,
        default="",
    )
    parser.add_argument("--account", dest="account", required=True)
    parser.add_argument("--warehouse", dest="warehouse", required=False)
    parser.add_argument("--database", dest="database", required=True)
    parser.add_argument("--schema", dest="schema", required=False, default="")
    parser.add_argument("--query", dest="query", required=True)
    parser.add_argument("--user-role", dest="user_role", required=False, default="")
    return parser.parse_args()


def main():
    try:
        args = get_args()
        client_args = {
            "username": args.username,
            "password": None if args.password == "" else args.password,
            "account": None if args.account == "" else args.account,
            "warehouse": args.warehouse,
            "schema": None if args.schema == "" else args.schema,
            "database": args.database,
            "rsa_key": None if args.private_key_path == "" else args.private_key_path,
            "role": None if args.user_role == "" else args.user_role,
        }
        private_key_passphrase = (
            None if args.private_key_passphrase == "" else args.private_key_passphrase
        )
        client = SnowflakeClient(**client_args)
        if client.rsa_key and not private_key_passphrase:
            logger.error("Private key passphrase is required when using a private key")
            sys.exit(client.EXIT_CODE_INVALID_CREDENTIALS)
        client.connect()
        client.execute_query(query=args.query)
    except ExitCodeException as e:
        logger.error(str(e.message))
        sys.exit(e.exit_code)
    else:
        logger.info("Successfully executed query")
        logger.info("Closing Connection")


if __name__ == "__main__":
    main()



import re
import psutil
import os
import pandas as pd
from datetime import datetime
from dask import dataframe as dd
from cryptography.hazmat.backends import default_backend
from cryptography.hazmat.primitives.asymmetric import rsa
from cryptography.hazmat.primitives.asymmetric import dsa
from cryptography.hazmat.primitives import serialization
from cryptography.hazmat.primitives.serialization.ssh import serialize_ssh_public_key
from typing import Dict, List, Optional, Tuple, Union
from copy import deepcopy
from random import random, randrange
from itertools import islice
from io import StringIO
from math import exp, log, floor, ceil


def _get_file_size(file: str) -> int:
    """Helper function to get the size of a given file

    Args:
        file: The desired file to grab the size of

    Returns: The size in bytes
    """
    return os.path.getsize(file)


def _get_memory() -> int:
    """Helper function to get the available memory

    Returns: The amount of memory in bytes

    """
    return psutil.virtual_memory().total


def _file_fits_in_memory(file: str) -> bool:
    """Helper function to show if a file can be read into memory

    Args:
        file: The desired file to be read in

    Returns: True if the file can fit in memory and False otherwise

    """
    if _get_file_size(file) < _get_memory():
        return True
    return False


def _decode_rsa(rsa_key: str):
    try:
        with open(rsa_key, "rb") as key:
            p_key = serialization.load_pem_private_key(
                key.read(),
                password=os.environ["SNOWFLAKE_PRIVATE_KEY_PASSPHRASE"].encode(),
                backend=default_backend(),
            )
        pkb = p_key.private_bytes(
            encoding=serialization.Encoding.DER,
            format=serialization.PrivateFormat.PKCS8,
            encryption_algorithm=serialization.NoEncryption(),
        )
        return pkb

    except Exception as e:
        return None


def map_snowflake_to_pandas(
    snowflake_data_types: Optional[Union[List[List], Dict[str, str]]]
) -> Union[Dict, None]:
    # TODO: modify this to accept a list of lists (old way) and a JSON representation of datatypes (new way)
    """Helper function to map a snowflake data type to the associated pandas data type

    Args:
        List

    Returns:
        dict | None: Dict where the key is the field name and the value is the pandas data type
    """
    if snowflake_data_types is None:
        return None
    snowflake_to_pandas = {
        "BOOLEAN": "bool",
        "TINYINT": "int8",
        "SMALLINT": "int16",
        "INTEGER": "int32",
        "INT": "int32",
        "BIGINT": "int64",
        "FLOAT": "float32",
        "DOUBLE": "float64",
        "DECIMAL": "float64",
        "NUMERIC": "float64",
        "NUMBER": "float64",
        "REAL": "float32",
        "DATE": "datetime64[ns]",
        "TIME": "datetime64[ns]",
        "DATETIME": "datetime64[ns]",
        "TIMESTAMP": "datetime64[ns]",
        "VARCHAR": "object",
        "NVARCHAR": "object",
        "CHAR": "object",
        "NCHAR": "object",
        "BINARY": "object",
        "VARBINARY": "object",
        "STRING": "object",
    }

    pandas_dtypes = {}
    if isinstance(snowflake_data_types, list):
        for item in snowflake_data_types:
            field = item[0]
            dtype = item[1]
            try:
                converted = snowflake_to_pandas[str(dtype).upper()]
                if converted is None:
                    raise Exception(
                        f"Invalid datatypes: the datatype {dtype} is not a recognized snowflake datatype"
                    )

                pandas_dtypes[field] = converted
            except KeyError as e:
                raise Exception(
                    f"Invalid datatype: the datatype {dtype} is not a recognized snowflake datatype"
                )
    elif isinstance(snowflake_data_types, dict):
        for field, dtype in snowflake_data_types.items():
            try:
                converted = snowflake_to_pandas[str(dtype).upper()]
                if converted is None:
                    raise Exception(
                        f"Invalid datatypes: the datatype {dtype} is not a recognized snowflake datatype"
                    )

                pandas_dtypes[field] = converted
            except KeyError as e:
                raise Exception(
                    f"Invalid datatype: the datatype {dtype} is not a recognized snowflake datatype"
                )
    else:
        raise Exception(
            "Unsupported format. Supplied datatypes should be either an array of arrays, or JSON "
        )

    return pandas_dtypes


def get_pandas_dates(pandas_datatypes: dict) -> tuple:
    dates = []
    new_dict = deepcopy(pandas_datatypes)
    for k, v in pandas_datatypes.items():
        if v in ["datetime64[ns]", "datetime64"]:
            dates.append(k)
            del new_dict[k]
    return (dates, new_dict) if dates else (None, new_dict)


def read_file(
    file: str, snowflake_dtypes: Union[List, None] = None, file_type: str = "csv"
) -> pd.DataFrame:
    """Helper function to read in a file to a pandas dataframe. This will be build out in the future to allow for more file types like parquet, arrow, tsv, etc.
    Args:
        file (str): The file to be read in as a dataframe
        pandas_dtypes (Union[Dict,None]): The optional dictionary of pandas data types to be used when reading in the file. Defaults to None
        file_type (str): The type of file to be read in. Defaults to 'csv'

    Returns:
        pd.DataFrame: The dataframe output of the file
    """
    if snowflake_dtypes:
        pandas_dtypes = map_snowflake_to_pandas(snowflake_dtypes)
        dates, pandas_dtypes = get_pandas_dates(
            pandas_dtypes
        )  # get the dates to be parsed

    else:
        pandas_dtypes = None
        dates = None
    if _file_fits_in_memory(file):
        # if it fits in memory
        if file_type == "csv":  # alwys true for now, will be augmented in the future
            if dates:
                df = pd.read_csv(
                    file,
                    dtype=pandas_dtypes,
                    parse_dates=dates,
                    date_parser=lambda x: pd.to_datetime(x).tz_localize("UTC"),
                )
            else:
                df = pd.read_csv(file, dtype=pandas_dtypes)
    else:
        # if the file is larger than memory
        if dates:
            df = dd.read_csv(
                file,
                dtype=pandas_dtypes,
                parse_dates=dates,
                date_parser=lambda x: pd.to_datetime(x).tz_localize("UTC"),
            )
        else:
            df = dd.read_csv("file", dtype=pandas_dtypes)
    return df


def reservoir_sample(iterable, k=1):
    """Select k items uniformly from iterable.
    Returns the whole population if there are k or fewer items

    """
    iterator = iter(iterable)
    values = list(islice(iterator, k))

    W = exp(log(random()) / k)
    while True:
        # skip is geometrically distributed
        skip = floor(log(random()) / log(1 - W))
        selection = list(islice(iterator, skip, skip + 1))
        if selection:
            values[randrange(k)] = selection[0]
            W *= exp(log(random()) / k)
        else:
            return values


def _parse_dates(df: pd.DataFrame) -> pd.DataFrame:
    for col in df.columns:
        # Check if the column contains date-like values
        if df[col].apply(lambda x: bool(re.match(r"\d{4}-\d{2}-\d{2}", str(x)))).all():
            df[col] = pd.to_datetime(df[col])

    return df


def infer_schema(file_name: str, k=10000) -> Dict[str, str]:
    """Randomly samples a csv file and infers the schema based off of the k rows sampled

    Args:
        file_name (str): The file to be sampled
        k (int, optional): The number of rows to sample. Defaults to 10000.

    Returns:
        dict: The dictionary of inferred pandas datatypes data types
    """
    if isinstance(file_name, list):
        dataframes = []
        n_files = len(file_name)
        rows_per_file = ceil(k / n_files)
        for f in file_name:
            file_path = f
            with open(file_path, "r") as f:
                header = next(f)
                result = [header] + reservoir_sample(f, rows_per_file)
            df = pd.read_csv(StringIO("".join(result)))
            dataframes.append(df)
        merged = pd.concat(dataframes, axis=0, ignore_index=True)
        return merged.dtypes.to_dict()
    else:
        with open(file_name, "r") as f:
            header = next(f)
            result = [header] + reservoir_sample(f, k)
        df = pd.read_csv(StringIO("".join(result)), parse_dates=True)
        df = _parse_dates(df)  # parse the date and datetime fields
        return {k: str(v) for k, v in df.dtypes.to_dict().items()}


def map_pandas_to_snowflake(data_type_dict: Dict[str, str]) -> Dict[str, str]:
    """Helper function to map the the pandas data types to the snowflake data types

    Args:
        pandas_dtypes (dict): Dictionary of the pandas data types

    Returns:
        dict: Dictionary of the mapped snowflake data types
    """
    snowflake_data_types = {
        "int64": "NUMBER",
        "float64": "FLOAT",
        "object": "STRING",
        "bool": "BOOLEAN",
        "datetime64[ns]": "TIMESTAMP",
        "timedelta64[ns]": "TIME",
    }

    snowflake_columns = {}
    for column_name, pandas_data_type in data_type_dict.items():
        snowflake_columns[column_name] = snowflake_data_types.get(
            str(pandas_data_type), "STRING"
        )

    return snowflake_columns



import os
import sys
from shipyard_snowflake import SnowflakeClient
from shipyard_templates import ExitCodeException, ShipyardLogger

logger = ShipyardLogger.get_logger()


def main():
    snowflake = SnowflakeClient(
        username=os.getenv("SNOWFLAKE_USERNAME"),
        password=os.getenv("SNOWFLAKE_PASSWORD"),
        account=os.getenv("SNOWFLAKE_ACCOUNT"),
    )
    try:
        conn = snowflake.connect()
        sys.exit(0)
    except ExitCodeException as e:
        logger.authtest(f"Credential test failed for the following reason: {e}")
        sys.exit(1)


if __name__ == "__main__":
    main()
